# Copyright: 2007-2012 Paul Obermeier (obermeier@poSoft.de)
# Distributed under BSD license.

if { $argc == 0 } {
    set distRootDir "C:/tmp/CawtDist"
} else {
    set distRootDir [lindex $argv 0]
}
set zipProg "C:/opt/7-Zip/7z.exe"

set cawtDir  [pwd]
set docDir   [file join $cawtDir "Documentation"]
set finalDir [file join $docDir  "Final"]
set testDir  [file join $cawtDir "TestPrograms"]

set auto_path [linsert $auto_path 0 $cawtDir [file join $cawtDir "Externals"]]

set retVal [catch {package require cawt} cawtVersion]

set distUserName "Cawt-$cawtVersion-User"
set distUserDir [file join $distRootDir $distUserName]

set distDevName "Cawt-$cawtVersion-Dev"
set distDevDir [file join $distRootDir $distDevName]

puts "Generating distribution of Cawt $cawtVersion"

puts "Cleaning test and doc directories ..."
cd $finalDir
file delete -force "CawtFigures"
cd $testDir
file delete -force "testOut"

puts "Generating user distribution in directory $distUserDir"
file mkdir $distRootDir
file delete -force $distUserDir
file mkdir $distUserDir

cd $cawtDir
foreach dir { "CawtCore" "CawtEarth" "CawtExcel" "CawtExplorer" "CawtMatlab" \
              "CawtPpt" "CawtWord" "TestPrograms" } {
    puts "    Copying CAWT package $dir ..."
    file copy [file join $cawtDir $dir] $distUserDir
}

file mkdir [file join $distUserDir "Externals"]
foreach dir { "Img" "tablelist" "twapi"} {
    puts "    Copying external package $dir ..."
    file copy [file join $cawtDir "Externals" $dir] [file join $distUserDir "Externals"]
}
file copy pkgIndex.tcl $distUserDir

file copy $finalDir $distUserDir
cd $distUserDir
file rename "Final" "Documentation"

proc PackDir { dir zipName } {
    puts "    Generating $zipName from directory $dir ..."
    file delete -force $zipName

    exec $::zipProg a -tzip ${dir}.zip $dir -mx5
    file rename -force ${dir}.zip $zipName
}

cd $distRootDir
PackDir $distUserName "$distUserName.zip"

puts "Generating developer distribution in directory $distDevDir"
file copy [file join $cawtDir] $distRootDir
cd $distRootDir
file rename "Cawt" $distDevName

# Remove obsolete files and directories
cd $distDevName
file delete -force "_FOSSIL_"
file delete -force "Applications"

cd $distRootDir
PackDir $distDevName "$distDevName.zip"

puts "Done"
